--  2002 Peter Thiemann
module Style (Style(..), using) where

import HTMLMonad hiding (map, div, head, span)
import List

infixl 5 :^:
infix 9 :=:

-- |First class style elements. Intended for CSS2.
data Style 
  = NoStyle
  | String :=: String	    -- ^primitive style @name :=: value@
  | Style  :^: Style	    -- ^combine two styles
  | Named String	    -- ^reference to style sheet

instance Show Style where
  showsPrec n NoStyle = id
  showsPrec n (s1 :=: s2) = showString s1 . showString ": " . showString s2
  showsPrec n (s1 :^: s2) = shows s1 . showString "; " . shows s2
  showsPrec n (Named s) = id

extractNamed :: Style -> [String]
extractNamed NoStyle   = []
extractNamed (Named n) = [n]
extractNamed (_ :=: _) = []
extractNamed (s1 :^: s2) = 
  nub (extractNamed s1 ++ extractNamed s2)

attach :: (Monad m) => Style -> WithHTML x m ()
attach sty = 
  attr "style" (show sty) ##
  case extractNamed sty of
    [n] -> attr "class" n
    _   -> empty

-- |Attaches a style to a field constructor.
using :: (Monad m)
	=> Style
	-> (WithHTML x m b -> c)	-- ^a field constructor, typically 'HTMLField'
	-> WithHTML x m b -> c		-- ^styled version of this constructor
using style elem x = elem (x ## attach style)
