--  2001 Peter Thiemann
module HTMLBase 
  (ATTR_(), attr_, attr_name, attr_value
  ,ELEMENT_(), element_, empty_, cdata_, comment_, doctype_
  ,CDATA_OPTIONS(..)
  ,add_, add_attr_
  ,get_attrs_
  ,BT(..)
  ,element_S, empty_S, cdata_S, comment_S, doctype_S
  ,element_T, empty_T, cdata_T, comment_T, doctype_T
  ,attr_SS, attr_SD, attr_TS, attr_TD
  ,showTemplatified)
where

import Char
import SHA1
import qualified Base32

data BT = STATIC | DYNAMIC | TOPLEVEL

-- untyped layer
-- attributes

data ATTR_ =
  ATTR_ { attr_BT :: BT
	, attr_value_BT :: BT
	, attr_name :: String
        , attr_value :: String
	}

attr_   = ATTR_ DYNAMIC  DYNAMIC
attr_SS = ATTR_ STATIC   STATIC
attr_SD = ATTR_ STATIC   DYNAMIC
attr_TS = ATTR_ TOPLEVEL STATIC
attr_TD = ATTR_ TOPLEVEL DYNAMIC

-- elements

data ELEMENT_ =
    ELEMENT_ { elem_BT :: BT
	     , tag :: String
    	     , attrs :: [ATTR_]
	     , elems :: [ELEMENT_]
	     }
  | EMPTY_   { elem_BT :: BT
	     , tag :: String
  	     , attrs :: [ATTR_]
	     }
  | CDATA_   { elem_BT :: BT
	     , elem_cdata :: String
	     }
  | COMMENT_ { elem_BT :: BT
	     , elem_comment :: String
	     }
  | DOCTYPE_ { elem_BT :: BT
	     , doctype :: [String]
  	     , elems :: [ELEMENT_]
	     }

data CDATA_OPTIONS = CDATA_ENCODED | CDATA_FORMATTED
  deriving (Eq)

element_  = ELEMENT_ DYNAMIC
element_S = ELEMENT_ STATIC
element_T = ELEMENT_ TOPLEVEL
empty_    = EMPTY_ DYNAMIC
empty_S   = EMPTY_ STATIC
empty_T   = EMPTY_ TOPLEVEL
makeEncoder opt = format . encode
  where format | CDATA_FORMATTED `elem` opt = id
	       | otherwise = htmlFormat
	encode | CDATA_ENCODED `elem` opt = id
	       | otherwise = htmlEncode
cdata_  opt = CDATA_ DYNAMIC . makeEncoder opt
cdata_S opt = CDATA_ STATIC . makeEncoder opt
cdata_T opt = CDATA_ TOPLEVEL . makeEncoder opt
comment_  = COMMENT_ DYNAMIC
comment_S = COMMENT_ STATIC
comment_T = COMMENT_ TOPLEVEL
doctype_  = DOCTYPE_ DYNAMIC
doctype_S = DOCTYPE_ STATIC
doctype_T = DOCTYPE_ TOPLEVEL

add_ e_ e'_ = e_ { elems = e'_ : elems e_}
add_attr_ e_ att = 
  let nameOfAtt = attr_name att
      all_attrs = attrs e_
      f [] = Nothing
      f (att' : attrs) = if attr_name att' == nameOfAtt 
      			 then return (att : attrs)
			 else f attrs >>= \ attrs' -> return (att' : attrs')
      new_attrs = case f all_attrs of
		    Nothing -> att : all_attrs
		    Just attrs -> attrs
  in  e_ { attrs = new_attrs }

get_attrs_ = attrs

-- show functions

instance Show ATTR_ where
  showsPrec i = shows_attribute
  showList    = shows_attributes

shows_attributes :: [ATTR_] -> ShowS
shows_attributes atts = foldr (.) id (map shows_attribute atts)

shows_attribute :: ATTR_ -> ShowS
shows_attribute a =
  showChar ' ' . showString (attr_name a) .
  case attr_value a of
    "()" ->
      id
    str@('\"':_) ->
      showString "=\"" . showString (read str) . showString "\""
    str ->
      showString "=\"" . showString str . showString "\""

instance Show ELEMENT_ where
  showsPrec i = shows_element
  showList = shows_elements

shows_elements :: [ELEMENT_] -> ShowS
shows_elements elts = foldr (.) id (reverse (map shows_element elts))

shows_element :: ELEMENT_ -> ShowS
shows_element (EMPTY_ bt tag atts) =
  showChar '<' . showString tag . shows atts . showString "\n/>"
shows_element (ELEMENT_ bt tag atts elts) =
  showChar '<' . showString tag . shows atts . showChar '>'
  . shows_elements elts
  . showString "</" . showString tag . showString "\n>"
shows_element (DOCTYPE_ bt strs elems) =
  showString "<!DOCTYPE" . 
  foldr (\str f -> showChar ' ' . showString str . f) id strs . showString "\n>" .
  showString "<!-- generated by WASH/HTML 0.11\n-->" .
  shows_elements elems .
  showChar '\n'
shows_element (CDATA_ bt str) =
  showString str
shows_element (COMMENT_ bt str) =
  showString "<!-- " . showString str . showString "\n-->"

htmlEncode "" = ""
htmlEncode (x:xs) = 
	case x of
	  '&' -> "&amp;" ++ htmlEncode xs
	  '<' -> "&lt;"  ++ htmlEncode xs
	  '>' -> "&gt;"  ++ htmlEncode xs
	  '\"' -> "&quot;" ++ htmlEncode xs
	  _ -> x : htmlEncode xs

htmlFormat s = s
{--htmlFormat "" = ""
htmlFormat (x:xs) =
	if isSpace x then
	  "\n " ++ htmlFormat (dropWhile isSpace xs)
	else
	  x : htmlFormat xs	  
--}
----------------------------------------------------------------------
-- template generation
----------------------------------------------------------------------
showTemplatified :: ELEMENT_ -> ShowS
showTemplatified = showTemplate . analyze

showTemplate :: (Template, Mt) -> ShowS
showTemplate (tis, mt) =
  let ts = closed mt in
  showDefinitions (length ts) ts .
  showString "<script name=\"BODY\">\n" .
  showActualArgs (actuals mt) .
  showString " (document)\n" .
  showString "</script>\n"

showDefinitions n [] =
  let scriptName = "STRINGBASE"
  in
  showString "<script name=\"" . showString scriptName . showString "\">\n" .
  showString "function S0 (str) {\n" .
  showString "return function (out) { out.write (str); }\n" .
  showString "}\n" .
  showString "</script>\n"
    
showDefinitions n ((nformals, t1) :ts) =
  let scriptName = Base32.encode (sha1 (show t1))
  in
  showString "<script name=\"" . showString scriptName . showString "\">\n" .
  showString "function O" . shows n . 
  showString "(" . showFormals nformals . showString ")\n" .
  showString "{\n" .
  showString "return function (out) {\n" .
  showBody t1 .
  showString "}\n" .
  showString "}\n" .
  showString "</script>\n" .
  showDefinitions (n-1) ts
  
showBody [] = 
  id
showBody (ti:tis) =
  showBody tis .
  showStatement ti .
  showString ";\n"

showStatement (TOut str) =
  showString "out.write (\"" .
  showString str .					    -- JS String quoting!
  showString "\")"
showStatement (TVar n) =
  showString "v" . shows n . showString " (out)"
showStatement (TCall fname args) =
  error "this should not happen"

showFormals 0 =
  id
showFormals n =
  (if n>1 then showFormals (n-1) . showString ", " else id) .
  showString "v" . shows n

showActuals [] =
  id
showActuals [ti] = 
  showActual ti
showActuals (ti:tis) =
  showActual ti .
  showString ", " .
  showActuals tis

showActual (TOut str) =
  showString "S0 (\"" .
  showString str .
  showString "\") "
showActual (TVar n) =
  error "this should not happen"
showActual (TCall fname args) =
  showString "O" . shows fname . showString " (" .
  showActualArgs args .
  showString ") "

showActualArgs [] = 
  id
showActualArgs [arg] =
  showActualArg arg
showActualArgs (arg: args) =
  showActualArgs args .
  showString ", " .
  showActualArg arg

showActualArg [x] =
  showActual x
showActualArg xs = 
  showString "function (out) { " .
  foldr g id xs .
  showString "}"
  where g x showrest = showActual x . showString " (out); " . showrest
	  


----------------------------------------------------------------------
-- template analysis
----------------------------------------------------------------------

analyze :: ELEMENT_ -> (Template, Mt)
analyze e = unM (collect e STATIC []) mt0

data Mt = Mt
    { open :: Templates
    , closed :: [(Int, Template)]
    , dynamics :: [Templates]
    , actuals :: Templates
    , count :: Int
    }
    deriving Show

mt0 = Mt { open = []
	 , closed = []
	 , dynamics = []
	 , actuals = []
	 , count = 0
	 }

type Templates = [Template]

type Template = [TemplateItem]

data TemplateItem
  = TOut String
  | TVar Int
  | TCall Int [Template]
  deriving Show

tout :: String -> [TemplateItem] -> [TemplateItem]
tout s (TOut s' : rest) = (TOut (s'++s) : rest) 
tout s tis = TOut s : tis

data M a = M { unM :: Mt -> (a, Mt) }

instance Monad M where
  return a = M (\t -> (a, t))
  m >>= f  = M (\t -> let (a,t') = unM m t in unM (f a) t')

pushOpen :: Template -> M ()
pushOpen t = M (\mt -> ((), mt { open = t : open mt
			       , dynamics = actuals mt : dynamics mt 
			       , actuals = []
			       } ))
popOpen :: M Template
popOpen = M (\mt -> (head (open mt), mt { open = tail (open mt)
					, actuals = head (dynamics mt)
					, dynamics = tail (dynamics mt)
					}))

pushClosed :: Template -> M Int
pushClosed tis = M (\mt -> (length (closed mt) + 1
			   ,mt { closed = (length (actuals mt), tis) : closed mt }))

pushActuals :: Template -> M Int
pushActuals tis = M (\mt -> (length (actuals mt) + 1, mt { actuals = tis : actuals mt }))

getActuals :: M Templates
getActuals = M (\mt -> (actuals mt, mt))

get :: (Mt -> x) -> M x
get f = M (\mt -> (f mt, mt))

-- |collect takes an element, a list of open templates, a list of finished
-- templates, and returns a pair (open templates, finished templates).
collect :: ELEMENT_ -> BT -> Template -> M Template
collect (EMPTY_ bt tag atts) cbt cur =
  let t1 = TOut ('<' : tag)
      t2 = tout "/>"
  in
  case bt of 
    TOPLEVEL -> 
      do pushOpen cur
	 ts <- collectAttrs atts STATIC [t1]
	 fname <- pushClosed (t2 ts)
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
	 
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do ts <- collectAttrs atts DYNAMIC [t1]
	     vname <- pushActuals (t2 ts)
	     return (TVar vname : cur)
	DYNAMIC ->
	  do ts <- collectAttrs atts DYNAMIC (t1 : cur)
	     return (t2 ts)
    STATIC ->
      do ts <- collectAttrs atts STATIC (t1 : cur)
	 return (t2 ts)

collect (ELEMENT_ bt tag atts elems) cbt cur =
  let t1 = TOut ('<' : tag)
      t1' = tout ('<' : tag)
      t2 = tout ">"
      t3 = tout ("</" ++ tag ++ ">")
  in
  case bt of 
    TOPLEVEL -> 
      do pushOpen cur
	 ts1 <- collectAttrs atts STATIC [t1]
	 ts2 <- collectElems elems STATIC (t2 ts1)
	 fname <- pushClosed (t3 ts2)
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
	 
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do ts1 <- collectAttrs atts DYNAMIC [t1]
	     ts2 <- collectElems elems DYNAMIC (t2 ts1)
	     vname <- pushActuals (t3 ts2)
	     return (TVar vname : cur)
	DYNAMIC ->
	  do ts1 <- collectAttrs atts DYNAMIC (t1' cur)
	     ts2 <- collectElems elems DYNAMIC (t2 ts1)
	     return (t3 ts2)
    STATIC ->
      do ts1 <- collectAttrs atts STATIC (t1' cur)
	 ts2 <- collectElems elems STATIC (t2 ts1)
	 return (t3 ts2)

collect (DOCTYPE_ bt strs elems) cbt cur =
  let t1 = TOut ("<!DOCTYPE" ++
		 (foldr (\str f -> showChar ' ' . showString str . f) id strs .
		  showString ">")
		 "<!-- generated by WASH/HTML 0.11\n-->")
  in
  case bt of 
    TOPLEVEL -> 
      do pushOpen cur
	 ts2 <- collectElems elems STATIC [t1]
	 fname <- pushClosed ts2
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
	 
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do ts2 <- collectElems elems DYNAMIC [t1]
	     vname <- pushActuals ts2
	     return (TVar vname : cur)
	DYNAMIC ->
	  do collectElems elems DYNAMIC (t1 : cur)
    STATIC ->
      do collectElems elems STATIC (t1 : cur)

collect (CDATA_ bt str) cbt cur =
  let t0 = TOut str
      t0' = tout str in
  case bt of 
    TOPLEVEL -> 
      do pushOpen cur
	 fname <- pushClosed [t0]
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do vname <- pushActuals [t0]
	     return (TVar vname : cur)
	DYNAMIC ->
	  do return [t0]
    STATIC ->
      do return (t0' cur)

collect (COMMENT_ bt str) cbt cur =
  let t0 = TOut str
      t0' = tout str in
  case bt of 
    TOPLEVEL -> 
      do pushOpen cur
	 fname <- pushClosed [t0]
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do vname <- pushActuals [t0]
	     return (TVar vname : cur)
	DYNAMIC ->
	  do return [t0]
    STATIC ->
      do return (t0' cur)

collectElems [] cbt cur =
  return cur
collectElems (at:ats) cbt cur =
  do ts <- collect at cbt cur
     collectElems ats cbt ts

collectAttrs [] cbt cur =
  return cur
collectAttrs (at:ats) cbt cur =
  do ts <- collectAttr at cbt cur
     collectAttrs ats cbt ts

collectAttr
  ATTR_ { attr_BT = bt
	, attr_value_BT = vbt
	, attr_name = aname
        , attr_value = aval
	} 
  cbt
  cur = let { t1 = TOut (' ' : aname ++ "=\"") ; t2 = TOut "\"";
	      t1' = tout (' ' : aname ++ "=\"") ; t2' = tout "\""; } in
  case bt of
    TOPLEVEL ->
      do pushOpen cur
	 ts <- collectAttrValue aval vbt STATIC [t1]
	 fname <- pushClosed (t2' ts)
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do ts <- collectAttrValue aval vbt DYNAMIC [t1]
	     vname <- pushActuals (t2' ts)
	     return (TVar vname : cur)
	DYNAMIC ->
	  do ts <- collectAttrValue aval vbt DYNAMIC (t1' cur)
	     return (t2' ts)
    STATIC ->
      do ts <- collectAttrValue aval vbt STATIC (t1' cur)
	 return (t2' ts)

collectAttrValue aval bt cbt cur =
  let t0 = TOut aval
      t0' = tout aval in
  case bt of
    TOPLEVEL ->
      do pushOpen cur
	 fname <- pushClosed [t0]
	 parms <- get actuals
	 cur' <- popOpen
	 vname <- pushActuals [TCall fname parms]
	 return (TVar vname : cur')
    DYNAMIC ->
      case cbt of
	STATIC ->
	  do vname <- pushActuals [t0]
	     return (TVar vname : cur)
	DYNAMIC ->
	  do return [t0]
    STATIC ->
      do return (t0' cur)
