module WASHExpression where

import WASHFlags
import qualified WASHUtil
import WASHData

code :: FLAGS -> [CodeFrag] -> ShowS
code flags [] = id
code flags (x:xs) = code' flags x . code flags xs

code' :: FLAGS -> CodeFrag -> ShowS
code' flags (HFrag h) = 
  showString h
code' flags (EFrag e) =
  element flags e
code' flags (CFrag cnts) =
  contents flags cnts
code' flags (AFrag attrs) =
  showChar '(' .
  WASHUtil.itemList (attribute flags) "empty" " >> " attrs .
  showChar ')'
code' flags _ = error "Unknown type: code"

showMode :: Mode -> ShowS
showMode V = id
showMode S = showString "_T"
showMode F = showString "_S"

element :: FLAGS -> Element -> ShowS
element flags (Element mode nm ats cnt et) =
  showString nm .
  (if generateBT flags then showMode mode else id) .
  showString " (" .
  attributes flags ats .
  contents flags cnt .
  showString ")"

contents :: FLAGS -> [Content] -> ShowS
contents flags cnts =
  showString "(do { " .
  WASHUtil.itemList (content flags) "empty" "; " (WASHUtil.normalize cnts) .
  showString ";})"

content :: FLAGS -> Content -> ShowS
content flags (CElement elem)  = 
  element flags elem
content flags (CText txt) =
  text flags txt
content flags (CCode c) =
  code flags c
content flags c = 
  error $ "Unknown type: content -- " ++ (show c)

text :: FLAGS -> Text -> ShowS
text flags txt =
  showString "rawtext" .
  (if generateBT flags then showMode (textMode txt) else id) .
  showChar ' ' .
  shows (textString txt)

attributes :: FLAGS -> [Attribute] -> ShowS
attributes flags atts = 
  f atts
    where
      f [] = id
      f (att:atts) = 
	attribute flags att .
	showString " >> " .
	f atts

attribute :: FLAGS -> Attribute -> ShowS
attribute flags (Attribute m n v) = 
  showString "(attr" .
  (if generateBT flags then (attrvalueBT m v) else id) .
  showChar ' ' .
  shows n . 
  showString " " .
  attrvalue v .
  showString ")"
attribute flags (AttrPattern pat) =
  showString "( " .
  showString pat .
  showString " )"
attribute flags a = error $ "Unknown type: attribute -- " ++ (show a)

attrvalue :: AttrValue -> ShowS
attrvalue (AText t) = 
  shows t
attrvalue (ACode c) =
  showString "( " .
  showString c .
  showString " )"
attrvalue a = error $ "Unknown type: attrvalue -- " ++ (show a)

attrvalueBT :: Mode -> AttrValue -> ShowS
attrvalueBT V _ = id
attrvalueBT m (AText _) = showMode m . showChar 'S'
attrvalueBT m (ACode _) = showMode m . showChar 'D'
attrvalueBT m a = error $ "Unknown type: attrvalueBT -- " ++ (show a)
